@@ -1,17 +1,17 @@
 from logging import getLogger
-from typing import Any, Dict, List, Optional, Tuple
+from typing import Any, Callable, Dict, List, Optional, Tuple, Union
 
-from watchmen_model.admin import Topic
-from watchmen_model.common import DataPage
+from watchmen_model.admin import Factor, Topic
+from watchmen_model.common import DataPage, Storable
 from watchmen_storage import as_table_name, Entity, EntityColumnAggregateArithmetic, EntityDeleter, \
 	EntityDistinctValuesFinder, EntityFinder, EntityHelper, EntityId, EntityIdHelper, EntityList, \
-	EntityNotFoundException, EntityPager, EntityStraightAggregateColumn, EntityStraightColumn, \
+	EntityNotFoundException, EntityPager, EntityRow, EntityStraightAggregateColumn, EntityStraightColumn, \
 	EntityStraightValuesFinder, EntityUpdater, FreeAggregatePager, FreeAggregator, FreeFinder, FreePager, \
 	TooManyEntitiesFoundException, TopicDataStorageSPI, TransactionalStorageSPI, UnexpectedStorageException, \
 	UnsupportedStraightColumnException
 from watchmen_utilities import ArrayHelper, is_blank
 from .document_defs_mongo import find_document, register_document
-from .document_mongo import DOCUMENT_OBJECT_ID, MongoDocument
+from .document_mongo import MongoDocument
 from .engine_mongo import MongoConnection, MongoEngine
 from .sort_build import build_sort_for_statement
 from .topic_document_generate import build_to_trino_fields
@@ -41,7 +41,7 @@ class StorageMongoDB(TransactionalStorageSPI):
 		self.connection.begin()
 
 	def commit_and_close(self) -> None:
-		pass
+		self.close()
 
 	def rollback_and_close(self) -> None:
 		try:
@@ -67,6 +67,7 @@ class StorageMongoDB(TransactionalStorageSPI):
 		document = self.find_document(helper.name)
 		entity = helper.shaper.serialize(one)
 		entity = document.copy_id_column_to_object_id(entity)
+		entity = document.change_date_to_datetime(entity)
 		self.connection.insert_one(document, entity)
 
 	def insert_all(self, data: List[Entity], helper: EntityHelper) -> None:
@@ -88,7 +89,7 @@ class StorageMongoDB(TransactionalStorageSPI):
 			name=updater.name,
 			shaper=updater.shaper,
 			criteria=updater.criteria,
-			distinctColumnNames=[DOCUMENT_OBJECT_ID],
+			distinctColumnNames=['_id'],
 			distinctValueOnSingleColumn=False
 		))
 
@@ -99,8 +100,11 @@ class StorageMongoDB(TransactionalStorageSPI):
 			else:
 				raise EntityNotFoundException(f'Entity not found by updater[{updater}]')
 		elif should_update_count == 1:
+			# noinspection PyProtectedMember,PyUnresolvedReferences
 			updated_count = self.connection.update_by_id(
-				document, updater.update, str(entities[0][DOCUMENT_OBJECT_ID])).modified_count
+				document, document.change_date_to_datetime(updater.update),
+				self.get_id_from_entity(entities[0])
+			).matched_count
 			if updated_count == 0:
 				# might be removed by another session
 				if peace_when_zero:
@@ -117,7 +121,7 @@ class StorageMongoDB(TransactionalStorageSPI):
 			name=updater.name,
 			shaper=updater.shaper,
 			criteria=updater.criteria,
-			distinctColumnNames=[DOCUMENT_OBJECT_ID],
+			distinctColumnNames=['_id'],
 			distinctValueOnSingleColumn=False
 		))
 
@@ -125,10 +129,13 @@ class StorageMongoDB(TransactionalStorageSPI):
 		if should_update_count == 0:
 			return None
 		elif should_update_count == 1:
-			object_id = str(entities[0][DOCUMENT_OBJECT_ID])
+			# noinspection PyProtectedMember,PyUnresolvedReferences
+			object_id = self.get_id_from_entity(entities[0])
 			entity = self.connection.find_by_id(document, object_id)
 			if entity is not None:
-				updated_count = self.connection.update_by_id(document, updater.update, object_id).modified_count
+				updated_count = self.connection.update_by_id(
+					document, document.change_date_to_datetime(updater.update), object_id
+				).modified_count
 				if updated_count == 0:
 					# might be removed by another session
 					return None
@@ -142,7 +149,9 @@ class StorageMongoDB(TransactionalStorageSPI):
 	def update(self, updater: EntityUpdater) -> int:
 		document = self.find_document(updater.name)
 		where = build_criteria_for_statement([document], updater.criteria)
-		return self.connection.update_many(document, updater.update, where).modified_count
+		return self.connection.update_many(
+			document, document.change_date_to_datetime(updater.update), where
+		).modified_count
 
 	# noinspection DuplicatedCode
 	def update_and_pull(self, updater: EntityUpdater) -> EntityList:
@@ -180,7 +189,7 @@ class StorageMongoDB(TransactionalStorageSPI):
 			name=deleter.name,
 			shaper=deleter.shaper,
 			criteria=deleter.criteria,
-			distinctColumnNames=[DOCUMENT_OBJECT_ID],
+			distinctColumnNames=['_id'],
 			distinctValueOnSingleColumn=False
 		))
 
@@ -188,7 +197,8 @@ class StorageMongoDB(TransactionalStorageSPI):
 		if should_delete_count == 0:
 			raise EntityNotFoundException(f'Entity not found by deleter[{deleter}].')
 		elif should_delete_count == 1:
-			deleted_count = self.connection.delete_by_id(document, str(entities[0][DOCUMENT_OBJECT_ID])).deleted_count
+			# noinspection PyProtectedMember,PyUnresolvedReferences
+			deleted_count = self.connection.delete_by_id(document, self.get_id_from_entity(entities[0])).deleted_count
 			if deleted_count == 0:
 				raise EntityNotFoundException(f'Entity not found by deleter[{deleter}].')
 			return deleted_count
@@ -201,7 +211,7 @@ class StorageMongoDB(TransactionalStorageSPI):
 			name=deleter.name,
 			shaper=deleter.shaper,
 			criteria=deleter.criteria,
-			distinctColumnNames=[DOCUMENT_OBJECT_ID],
+			distinctColumnNames=['_id'],
 			distinctValueOnSingleColumn=False
 		))
 
@@ -209,7 +219,8 @@ class StorageMongoDB(TransactionalStorageSPI):
 		if should_delete_count == 0:
 			return None
 		elif should_delete_count == 1:
-			object_id = str(entities[0][DOCUMENT_OBJECT_ID])
+			# noinspection PyProtectedMember,PyUnresolvedReferences
+			object_id = self.get_id_from_entity(entities[0])
 			entity = self.connection.find_by_id(document, object_id)
 			if entity is not None:
 				deleted_count = self.connection.delete_by_id(document, object_id).deleted_count
@@ -246,8 +257,9 @@ class StorageMongoDB(TransactionalStorageSPI):
 
 	# noinspection PyMethodMayBeStatic
 	def remove_object_id(self, data: Dict[str, Any]) -> Dict[str, Any]:
-		if DOCUMENT_OBJECT_ID in data:
-			del data[DOCUMENT_OBJECT_ID]
+		if '_id' in data:
+			# noinspection PyProtectedMember,PyUnresolvedReferences
+			del data['_id']
 		return data
 
 	def find_by_id(self, entity_id: EntityId, helper: EntityIdHelper) -> Optional[Entity]:
@@ -297,9 +309,24 @@ class StorageMongoDB(TransactionalStorageSPI):
 		else:
 			results = self.connection.find_distinct(document, finder.distinctColumnNames[0], where)
 
+		def try_to_remove_object_id(data: Dict[str, Any]) -> Dict[str, Any]:
+			if '_id' not in finder.distinctColumnNames:
+				return self.remove_object_id(data)
+			else:
+				return data
+
+		def decorate_deserialize(deserialize: Callable[[EntityRow], Entity]) -> Callable[[EntityRow], Entity]:
+			def action(row: EntityRow) -> Entity:
+				entity = deserialize(row)
+				if '_id' in finder.distinctColumnNames:
+					entity = self.bind_id_to_entity(entity, row.get('_id'))
+				return entity
+
+			return action
+
 		return ArrayHelper(results) \
-			.map(self.remove_object_id) \
-			.map(finder.shaper.deserialize) \
+			.map(try_to_remove_object_id) \
+			.map(decorate_deserialize(finder.shaper.deserialize)) \
 			.to_list()
 
 	# noinspection PyMethodMayBeStatic
@@ -366,8 +393,8 @@ class StorageMongoDB(TransactionalStorageSPI):
 		straight_columns = finder.straightColumns
 		aggregate_columns = self.build_straight_aggregate_columns(straight_columns)
 		if len(aggregate_columns) == 0:
-			def add_column(columns: Dict[str, int], column_name: str) -> Dict[str, int]:
-				columns[column_name] = 1
+			def add_column(columns: Dict[str, int], column: EntityStraightColumn) -> Dict[str, int]:
+				columns[column.columnName] = 1
 				return columns
 
 			project = ArrayHelper(straight_columns).reduce(add_column, {})
@@ -377,11 +404,11 @@ class StorageMongoDB(TransactionalStorageSPI):
 		else:
 			non_aggregate_columns = self.build_straight_non_aggregate_columns(straight_columns)
 			group = {
-				DOCUMENT_OBJECT_ID: non_aggregate_columns,
+				'_id': non_aggregate_columns,
 				**aggregate_columns
 			}
 			project = {
-				DOCUMENT_OBJECT_ID: 0,
+				'_id': 0,
 				**self.build_straight_project_columns(straight_columns, non_aggregate_columns, aggregate_columns)
 			}
 			where = build_criteria_for_statement([document], finder.criteria)
@@ -453,6 +480,22 @@ class StorageMongoDB(TransactionalStorageSPI):
 		where = build_criteria_for_statement([document], finder.criteria)
 		return self.connection.count(document, where)
 
+	# noinspection PyMethodMayBeStatic
+	def bind_id_to_entity(self, entity: Union[Storable, Dict], id_: Any) -> Union[Storable, Dict]:
+		if isinstance(entity, dict):
+			entity['_id'] = id_
+		else:
+			entity._id = id_
+		return entity
+
+	# noinspection PyMethodMayBeStatic
+	def get_id_from_entity(self, entity: Union[Storable, Dict]) -> Union[str, int]:
+		if isinstance(entity, dict):
+			return entity['_id']
+		else:
+			# noinspection PyProtectedMember
+			return entity._id
+
 
 class TopicDataStorageMongoDB(StorageMongoDB, TopicDataStorageSPI):
 	def register_topic(self, topic: Topic) -> None:
@@ -466,8 +509,8 @@ class TopicDataStorageMongoDB(StorageMongoDB, TopicDataStorageSPI):
 		# update collection is unnecessary
 		pass
 
-	def drop_topic_entity(self, topic_name: str) -> None:
-		entity_name = as_table_name(topic_name)
+	def drop_topic_entity(self, topic: Topic) -> None:
+		entity_name = as_table_name(topic)
 		try:
 			self.connect()
 			self.connection.drop_collection(entity_name)
@@ -483,38 +526,46 @@ class TopicDataStorageMongoDB(StorageMongoDB, TopicDataStorageSPI):
 		document = self.find_document(helper.name)
 		self.connection.drop_collection(document.name)
 
+	def ask_synonym_factors(self, table_name: str) -> List[Factor]:
+		"""
+		not supported by mongo
+		"""
+		raise UnexpectedStorageException('Method[ask_synonym_factors] does not support by mongo storage.')
+
 	def is_free_find_supported(self) -> bool:
 		return False
 
 	def append_topic_to_trino(self, topic: Topic) -> None:
+		self.connect()
 		self.connection.insert_one(self.find_document('_schema'), {
 			'table': as_table_name(topic.name),
 			'fields': ArrayHelper(build_to_trino_fields(topic)).map(lambda x: x.to_dict()).to_list()
 		})
 
 	def drop_topic_from_trino(self, topic: Topic) -> None:
+		self.connect()
 		self.connection.delete_many(self.find_document('_schema'), {'table': as_table_name(topic.name)})
 
 	def free_find(self, finder: FreeFinder) -> List[Dict[str, Any]]:
 		"""
 		not supported by mongo
 		"""
-		raise UnexpectedStorageException(f'Method[free_find] does not support by mongo storage.')
+		raise UnexpectedStorageException('Method[free_find] does not support by mongo storage.')
 
 	def free_page(self, pager: FreePager) -> DataPage:
 		"""
 		not supported by mongo
 		"""
-		raise UnexpectedStorageException(f'Method[free_page] does not support by mongo storage.')
+		raise UnexpectedStorageException('Method[free_page] does not support by mongo storage.')
 
 	def free_aggregate_find(self, aggregator: FreeAggregator) -> List[Dict[str, Any]]:
 		"""
 		not supported by mongo
 		"""
-		raise UnexpectedStorageException(f'Method[free_aggregate_find] does not support by mongo storage.')
+		raise UnexpectedStorageException('Method[free_aggregate_find] does not support by mongo storage.')
 
 	def free_aggregate_page(self, pager: FreeAggregatePager) -> DataPage:
 		"""
 		not supported by mongo
 		"""
-		raise UnexpectedStorageException(f'Method[free_aggregate_page] does not support by mongo storage.')
+		raise UnexpectedStorageException('Method[free_aggregate_page] does not support by mongo storage.')
